declare module "sourced" {
	import {EventEmitter} from 'events';

	/**
	 * Extends native Error object
	 */
	export class EntityError extends Error {
		/**
		 * @param {string} message An optional error message.
		 * @param {object} constructor An optional constructor to pass to `captureStackTrace`. Default: `this`.
		 */
		constructor(message?: string, constructor?: any);
	}

	export class Entity extends EventEmitter {
		/**
		 * Helper function to automatically create a method that calls digest on the
		 * param provided. Use it to add methods that automatically call digest.
		 *
		 * @param  {Object} type  the entity class to which the method will be added.
		 * @param  {Function} fn  the actual function to be added.
		 * @example
		 *
		 *    Entity.digestMethod(Car, function clearSettings (param) {
		 *     // Note, `this` is the extended Entity.
		 *     this.settings.get(param.name).forEach((name, config) => {
		 *       config.sources.forEach((source) => {
		 *         source.remove();
		 *       });
		 *     });
		 *
		 *     return this.settings;
		 *    });
		 *
		 */
		public static digestMethod(type: string, fn: Function): any;

		/**
		 * An array of events to emit when a repository commits the entity.
		 */
		public eventsToEmit: string[];

		public newEvents: any[];

		/**
		 * Boolean to prevent emit, enqueue and digest from running during replay.
		 */
		public replaying: boolean;

		/**
		 * Holds the version of the latest snapshot for the entity.
		 */
		public snapshotVersion: number;

		/**
		 * Holds the event's timestamp in the entity.
		 */
		public timestamp: number;

		/**
		 * Holds the current version of the entity.
		 */
		public version: number;

		public constructor(snapshot?: any, events?: any[]);

		/**
		 * Will emit an event if not replaying.
		 */
		public emit(event: string, ...args: any[]): boolean;

		/**
		 * Adds an array of events to `eventsToEmit` which are then emitted when the repository commits the entity.
		 * This is used in leiu of using `emit` directly to ensure the event is fired only on a successful commit.
		 */
		public enqueue(...args: string[]): void;

		/**
		 * Digest a command with given data. This is called whenever you want to record
		 * a command into the events for the entity. If called during replay, this
		 * method does nothing.
		 *
		 * @param  {String} method  the name of the method/command you want to digest.
		 * @param  {Object} data    the data that should be passed to the replay.
		 */
		public digest(method: string, data: any): void;

		/**
		 * Merge a snapshot onto the entity.
		 *
		 * For every property passed in the snapshot, the value is deep-cloned and then
		 * merged into the instance through mergeProperty. See mergeProperty for details.
		 *
		 * @param  {Object} snapshot  snapshot object.
		 */
		public merge(snapshot: any): this;

		/**
		 * Merge a property onto the instance.
		 *
		 * Given a name and a value, mergeProperty checks first attempt to find the
		 * property in the mergeProperties map using the constructor name as key. If it
		 * is found and it is a function, the function is called. If it is NOT found
		 * we check if the property is an object. If so, we merge. If not, we simply
		 * assign the passed value to the instance.
		 *
		 * @param  {String} name   the name of the property being merged.
		 * @param  {Object} value  the value of the property being merged.
		 */
		public mergeProperty(name: string, value: any): void;

		/**
		 * Replay an array of events onto the instance.
		 *
		 * The goal here is to allow application of events without emitting, enqueueing
		 * nor digesting the replayed events. This is done by setting this.replaying to
		 * true which emit, enqueue and digest check for.
		 *
		 * If the method in the event being replayed exists in the instance, we call
		 * the mathod with the data in the event and set the version of the instance to
		 * the version of the event. If the method is not found, we attempt to parse the
		 * constructor to give a more descriptive error.
		 *
		 * @param  {Array} events  an array of events to be replayed.
		 * @throws {EntityError} if the event does not exist as a method on the Entity.
		 */
		public replay(events: string[]): void;

		/**
		 * Create a snapshot of the current state of the entity instance.
		 *
		 * Here the instance's snapshotVersion property is set to the current version,
		 * then the instance is deep-cloned and the clone is trimmed of the internal
		 * sourced attributes using trimSnapshot and returned.
		 *
		 * @returns  {Object}
		 */
		public snapshot(): any;

		/**
		 * Remove the internal sourced properties from the passed snapshot.
		 *
		 * Snapshots are to contain only entity data properties. This trims all other
		 * properties from the snapshot.
		 *
		 * @param  {Object} snapshot  the snapshot to be trimmed.
		 * @see Entity.prototype.snapshot
		 */
		public trimSnapshot(snapshot: any): any;

		/**
		 * Convenience function to store references to functions that should be run
		 * when merging a particular property.
		 *
		 * @param  {Object} ctor  the entity class to which the property->fn belongs to.
		 * @param  {String} name  the name of the property that holds the fn.
		 * @param  {Function} fn  the function to execute when merging the property.
		 * @example
		 *  function Wheel (status) {
		 *    this.status = status;
		 *  }
		 *
		 *  Wheel.prototype.go = function () {
		 *    this.status = 'going';
		 *  }
		 *
		 *  function Car () {
		 *    this.id = null;
		 *    this.wheel = new Wheel(); // for instantiating our default wheel, when we first 'new' up a Car
		 *
		 *    Entity.apply(this, arguments);
		 *  }
		 *
		 *  util.inherits(Car, Entity);
		 *
		 *  Entity.mergeProperty(Car, 'wheels', function (obj) {
		 *    this.wheel = new Wheel(); // for instantiating our wheel from saved values in a database
		 *  });
		 */
		public static mergeProperty(ctor: any, name: string, fn: Function): void
	}
}

declare module "sourced-repo-mongo" {
	import {EventEmitter} from 'events';
	import {Entity} from "sourced";

	export interface RepositoryOptions {
		indices?: string[];
		snapshotFrequency?: number;
	}

	export interface CommitOptions {
		forceSnapshot?: boolean;
	}

	export class Repository extends EventEmitter {
		entityType: any;

		constructor(ctor: any, options?: RepositoryOptions);

		commit(entity: Entity, options: CommitOptions, callback: Function): any;
		commit(entity: Entity, callback: Function): any;

		commitAll(entities: Entity[], options: CommitOptions, callback: Function): any;
		commitAll(entities: Entity[], callback: Function): any;

		get(id: string, callback: (err: Error, entity: Entity) => void): void;

		getAll(ids: string[], callback: Function): any;
	}
}
